/***********************************************************************
/
/  EXTERNAL BOUNDARY CLASS
/
/  written by: Greg Bryan
/  date:       November, 1994
/  modified1:
/
/  PURPOSE:
/
************************************************************************/
#ifndef _h_ExternalBoundary
#define _h_ExternalBoundary

#include "xml_macro.h"
#include <mpi.h>

/* forward declaration for grid class */
class grid;


class ExternalBoundary
{
 private:
  int  BoundaryRank;                      // This is the rank and dimension 
  int  BoundaryDimension[MAX_DIMENSION];  //  of the grid to which the boundary
                                          //  values apply
  int  NumberOfBaryonFields;              // Number of boundary fields
  int  BoundaryFieldType[MAX_NUMBER_OF_BARYON_FIELDS];           
                                          // Field types

  boundary_type *BoundaryType[MAX_NUMBER_OF_BARYON_FIELDS][MAX_DIMENSION][2];
                                          /* Type of boundary value:
                                              1 - reflecting
					      2 - outflow
					      3 - inflow
					      4 - periodic   */

  boundary_type ParticleBoundaryType;

  float *BoundaryValue[MAX_NUMBER_OF_BARYON_FIELDS][MAX_DIMENSION][2];  
					  // boundary values for inflow (3)

  friend class grid;

 public:
//
// Constructor (set rank and dims from TopGrid)
//
  ExternalBoundary();
//
// Prepare External Boundary (set dims, etc.) based on TopGrid in argument.
//
  int Prepare(grid *TopGrid);
//
// Checks if the External Boundary has been prepared
//
  int AmIPrepared() {return (BoundaryRank > 0) ? TRUE : FALSE;};
//
// Set one face of external boundaries to a constant value 
//  (Note: this is not suitable for setting inflow conditions as
//         BoundaryValue is not set).
//  Returns SUCCESS or FAIL.
//
  int InitializeExternalBoundaryFace(int Dimension, 
			      boundary_type LeftBoundaryType,
			      boundary_type RightBoundaryType, 
			      float LeftBoundaryValue[],
			      float RightBoundaryValue[]);
//
//  Initialize particle boundary conditions
//
  void InitializeExternalBoundaryParticles(boundary_type ParticleBoundary)
    {ParticleBoundaryType = ParticleBoundary;};
//
// Read an external boundary
//
  int ReadExternalBoundary(FILE *fptr);
//
// Write an external boundary
//
  int WriteExternalBoundary(FILE *fptr, char *hdfname);
#ifdef _HDF5_OUTPUT
  int WriteExternalBoundaryHDF5(FILE*, hid_t, char*);
#endif
#ifdef _MPI_IO_OUTPUT
  int WriteExternalBoundaryMPIO(FILE*, MPI_File, MPI_Offset*, char*);
#endif
//
// Given a pointer to a field and its field type, find the equivalent
//   field type in the list of boundary's and apply that boundary value/type.
//   Returns: 0 on failure
//
  int SetExternalBoundary(int FieldRank, int GridDims[], int GridOffset[],
                          int StartIndex[], int EndIndex[],
                          float *Field, int FieldType);
//
// This routine handle the boundary conditions for particles.  The conditions
//   are assumed to be the same as the mass field.
//
  int SetExternalBoundaryParticles(int FieldRank, int NumberOfParticles,
                                   float *Position[], float *Velocity[]);
//
// Finds and returns the indexes to commonly used physical quantities.
//
  int IdentifyPhysicalQuantities(int &DensNum, int &GENum, int &Vel1Num, 
                                 int &Vel2Num, int &Vel3Num, int &TENum);
//
/************************************************************************/
//
// WavePool test problem:
//  This routine sets up the inflow boundary conditions to model an inflowing
//   linear wave (from the left boundary).  See also WavePoolGlobalData.h.
//
  int SetWavePoolBoundary(float time);
//
// ShockPool test problem:
//  This routine sets up the inflow boundary conditions to model an inflowing
//   shock wave (from the left boundary).  See also ShockPoolGlobalData.h.
//
  int SetShockPoolBoundary(float time);
//
// DoubleMach problem:
//  This routine sets up the necessary inflow boundary conditions.
//
  int SetDoubleMachBoundary(float time, float CellLeftEdge[], 
                            float CellWidth[]);

};
#endif
