/*************************************************************************/
/**   File:         fuzzy_kmeans.c                                      **/
/**   Description:  Implementation of fuzzy c-means clustering          **/
/**                 algorithm and cluster validity function.            **/
/**   Author:  Brendan McCane                                           **/
/**            James Cook University of North Queensland.               **/
/**            Australia. email: mccane@cs.jcu.edu.au                   **/
/**   Copywrite, 1995, Brendan McCane                                   **/
/**                                                                     **/
/*************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <float.h>
#include <math.h>
#include "fuzzy_kmeans.h"

#define SQR(x) ((x)*(x))

extern double wtime(void);

/*----< multid_feuclid_dist() >----------------------------------------------*/
/* multi-dimensional spatial Euclid distance */
__inline
float multid_feuclid_dist(float *pt1,
                          float *pt2,
                          int    numdims)
{
    float ans=0.0;
    int i;

    for (i=0; i<numdims; i++)
        ans += SQR(pt1[i]-pt2[i]);

    ans = sqrt((double)ans);
    return(ans);
}

/*----< find_nearest_point() >-----------------------------------------------*/
__inline
int find_nearest_point(float  *pt,          /* [nfeatures] */
                       int     nfeatures,
                       float **pts,         /* [npts][nfeatures] */
                       int     npts)
{
    int index, i;
    float max_dist=FLT_MAX;

    /* find the cluster center id with min distance to pt */
    for (i=0; i<npts; i++) {
        float dist;
        dist = euclid_dist_2(pt, pts[i], nfeatures);  /* no need square root */
        if (dist < max_dist) {
            max_dist = dist;
            index    = i;
        }
    }
    return(index);
}

/* --------------------------------------------------------------------
 * Procedure: fuzzy_kmeans_cluster
 * Comments:  cluster elements into nclusters clusters.
 * -------------------------------------------------------------------- */

/*----< sum_fuzzy_members() >------------------------------------------------*/
__inline
void sum_fuzzy_members(float **feature,         /* [npoints][nfeatures] */
                       int     npoints,
                       float **cluster_centres, /* [nclusters] */
                       int     nclusters,
                       int     nfeatures,
                       float   fuzzyq,
                       float  *sum_points)      /* [npoints] */
{
    int   i, j;
    float distance;

    for (i=0; i<npoints; i++) { /* for every object */
        sum_points[i] = 0.0;
        for (j=0; j<nclusters; j++) { /* for each cluster center */
            /* square of Euclid distance */
            distance = euclid_dist_2(feature[i], cluster_centres[j], nfeatures);

            /* ??? */
            /* sum of all fuzzy distance to the cluster centers */
            sum_points[i] += pow(1.0/(distance+0.00001), 1.0/(fuzzyq-1));
        }
    }
}

#define RANDOM_MAX 2147483647

/*----< fuzzy_kmeans_cluster() >---------------------------------------------*/
float **fuzzy_kmeans_cluster(float **feature,    /* [npoints][nfeatures] */
                             int     nfeatures,
                             int     npoints,
                             int     nclusters,
                             float   fuzzyq,
                             float   delta_centres_threshold)
                                     /* is this the tolerate error ? */
{
    float **cluster_centres;
    float max_delta_centres=0.0;
    float *sum_points;
    int i, j, k, index, loop=0;
    double timing;

#ifdef _LOOP_K_CLUSTER_FIRST_
    float *membership, *new_centre;
    /* indicates degree of membership of pt in cluster */
    membership = (float*) malloc(npoints * sizeof(float));

    /* temp variable for storing the new centre of a cluster */
    new_centre = (float*) malloc(nfeatures * sizeof(float));
#else
    float  *sum;         /* [nclusters] */
    float **new_centre;  /* [nclusters][nfeatures] */

    /* need sum and new_centre[0] initial 0 */
    sum = (float*) calloc(nclusters, sizeof(float));
    new_centre    = (float**) malloc(nclusters *            sizeof(float*));
    new_centre[0] = (float*)  calloc(nclusters * nfeatures, sizeof(float));
    for (i=1; i<nclusters; i++)
        new_centre[i] = new_centre[i-1] + nfeatures;
#endif

    cluster_centres   = (float**) malloc(nclusters * sizeof(float*));
    cluster_centres[0]= (float*)  malloc(nclusters * nfeatures * sizeof(float));
    for (i=1; i<nclusters; i++)
        cluster_centres[i] = cluster_centres[i-1] + nfeatures;

    /* cluster vector: coords of each cluster centre */
    for (i=0; i<nclusters; i++) {
        /* randomly pick a cluster centre */
        int n = (int)random() % npoints;
        for (j=0; j<nfeatures; j++)
            cluster_centres[i][j] = feature[n][j];
/*
            cluster_centres[i][j] = feature[i*(int)(npoints/nclusters)][j]
                                  + random()/(float)RANDOM_MAX;
*/
    }

    /* sum_points: vector for sums of dists of each pt to all clusters */
    sum_points = (float*) malloc(npoints * sizeof(float));

    if (_debug) timing = wtime();
    do {
        max_delta_centres = 0.0;
        /* to improve efficiency without using bazillions of memory */
        /* calc the sum of distances from each pt to all clusters */
        sum_fuzzy_members(feature,         /* [npoints][nfeatures] */
                          npoints,
                          cluster_centres, /* [nclusters] */
                          nclusters,
                          nfeatures,
                          fuzzyq,
                          sum_points);     /* out: [npoints] */
#ifdef _LOOP_K_CLUSTER_FIRST_
        for (i=0; i<nclusters; i++) {
            float sum, delta_centre;
            /* calculate fuzzy memberships of each point */
            sum = 0;
            for (j=0; j<npoints; j++) {
                float sqd_dist;
                sqd_dist = euclid_dist_2(feature[j],
                                         cluster_centres[i],
                                         nfeatures);
                /* the possibility of object j belong to clustr i */
                membership[j] = pow(1.0/(sqd_dist+0.00001), 1.0/(fuzzyq-1)) /
                                (sum_points[j]+0.00001);
                sum += pow(membership[j], fuzzyq);
            }
            /* calculate new cluster centre */
            for (j=0; j<nfeatures; j++) {
                new_centre[j] = 0.0;
                for (k=0; k<npoints; k++)
                    new_centre[j] += pow(membership[k], fuzzyq) *
                                     feature[k][j];
                new_centre[j] /= sum;
            }
            
            /* distance between new and old centers */
            delta_centre = multid_feuclid_dist(cluster_centres[i],
                                               new_centre,
                                               nfeatures);
            for (j=0; j<nfeatures; j++)
                cluster_centres[i][j] = new_centre[j];

            /* find max of center changes among k centers */
            if (delta_centre>max_delta_centres)
                max_delta_centres = delta_centre;
        }
#else
        for (j=0; j<npoints; j++) {
            /* calculate fuzzy memberships of each point j to cluster i */
            for (i=0; i<nclusters; i++) {
                float sqd_dist, membership, pow_mem;
                sqd_dist = euclid_dist_2(feature[j],
                                         cluster_centres[i],
                                         nfeatures);
                /* indicates degree of membership of pt in cluster */
                /* the possibility of object j belong to clustr i */
                membership = pow(1.0/(sqd_dist+0.00001), 1.0/(fuzzyq-1)) /
                             (sum_points[j]+0.00001);
                membership = pow(membership, fuzzyq);
                sum[i] += membership;

                for (k=0; k<nfeatures; k++)
                    new_centre[i][k] += membership * feature[j][k];
            }
        }

        for (i=0; i<nclusters; i++) {
            float delta_centre;
            for (k=0; k<nfeatures; k++)
                new_centre[i][k] /= sum[i];
            sum[i] = 0.0; /* reset sum[] to 0 */
            
            /* distance between new and old centers */
            delta_centre = multid_feuclid_dist(cluster_centres[i],
                                               new_centre[i],
                                               nfeatures);
            for (j=0; j<nfeatures; j++) {
                cluster_centres[i][j] = new_centre[i][j];
                new_centre[i][j] = 0.0; /* reset new_centre[] to 0 */
            }

            /* find max of center changes among k centers */
            if (delta_centre > max_delta_centres)
                max_delta_centres = delta_centre;
        }
#endif
    } while (max_delta_centres > delta_centres_threshold && loop++ < 500);

    if (_debug) {
        timing = wtime() - timing;
        printf("nloops = %2d (T = %7.4f)",loop,timing);
    }
#ifdef _LOOP_K_CLUSTER_FIRST_
    free(membership);
    free(new_centre);
#else
    free(sum);
    free(new_centre[0]);
    free(new_centre);
#endif
    free(sum_points);

    return(cluster_centres);
}

/* --------------------------------------------------------------------
 * Procedure: fuzzy_validity
 * Comments:  Calculates a fuzzy validity criterion based on compactness
 *            and separation of clusters.
 *              From Xie and Beni, vol PAMI-13, no8, pp.841-847, August, 91.
 * -------------------------------------------------------------------- */

/*----< fuzzy_validity() >---------------------------------------------------*/
float fuzzy_validity(float **feature,         /* [npoints][nfeatures] */
                     int     nfeatures,
                     int     npoints,
                     float **cluster_centres, /* [nclusters][nfeatures] */
                     int     nclusters, 
                     float   fuzzyq)
{
    int    i, j, k;
    int   *numpoints_in_clusters;
    float *sum_points;
    float  valid_sum=0.0, min_dist=FLT_MAX;
    float  ret, penalty;

    if (nclusters<2) return(FLT_MAX);

    numpoints_in_clusters = (int*)calloc(nclusters, sizeof(int));
    /* calculate number of points in each cluster */
    for (i=0; i<npoints; i++) {
        int nearest_cluster;
        nearest_cluster = find_nearest_point(feature[i], 
                                             nfeatures, 
                                             cluster_centres, 
                                             nclusters);
        numpoints_in_clusters[nearest_cluster]++;
    }

    /* calculate function which penalises validity if the clusters have
       a small number of points in them. */
    penalty=0.0;
    for (i=0; i<nclusters; i++)
        penalty += 1.0/(1.0+(float)numpoints_in_clusters[i]);
    penalty /= (float)nclusters;
    free(numpoints_in_clusters);
/*
    feature = (float**)calloc(npoints, sizeof(float*));
    for (i=0; i<npoints; i++)
        feature[i] = (float*)calloc(nfeatures, sizeof(float));
*/

    /* sum_points: vector for sums of dists of each pt to all clusters */
    sum_points = (float*) malloc(npoints * sizeof(float));

    /* to improve efficiency without using bazillions of memory */
    /* calc the sum of distances from each pt to all clusters */
    sum_fuzzy_members(feature,
                      npoints,
                      cluster_centres,
                      nclusters,
                      nfeatures,
                      fuzzyq,
                      sum_points);   /* out: [npoints] */

#ifdef _LOOP_K_CLUSTER_FIRST_
    for (i=0; i<nclusters; i++) {
        /* calculate fuzzy memberships of each point */
        for (j=0; j<npoints; j++) {
            float sqd_dist, memb_fuzzyq, membership;
            sqd_dist = euclid_dist_2(feature[j],
                                     cluster_centres[i],
                                     nfeatures);
            membership = pow(1.0/(sqd_dist+0.00001), 1.0/(fuzzyq-1)) /
                         (sum_points[j]+0.00001);
            memb_fuzzyq = pow(membership, fuzzyq);

            valid_sum += sqd_dist*memb_fuzzyq;
        }
    }
#else
    for (j=0; j<npoints; j++) {
        /* calculate fuzzy memberships of each point */
        for (i=0; i<nclusters; i++) {
            float sqd_dist, membership;
            sqd_dist = euclid_dist_2(feature[j],
                                     cluster_centres[i],
                                     nfeatures);
            membership = pow(1.0/(sqd_dist+0.00001), 1.0/(fuzzyq-1)) /
                         (sum_points[j]+0.00001);
            membership = pow(membership, fuzzyq);

            valid_sum += sqd_dist * membership;
        }
    }
#endif
    free(sum_points);

    /* find minimum distance between any two cluster centres */
    /*                               ^^^^^^^  ???            */
#define _ORIGINAL_
#ifdef _ORIGINAL_
    min_dist = 0.0;
    for (i=0; i<nclusters-1; i++) {
        float this_cluster_min_dist=FLT_MAX;
        for (j=i+1; j<nclusters; j++) {
            /* ??? these 2 loops are not for ANY TWO ??? */
            float tmp_dist;
            tmp_dist = multid_feuclid_dist(cluster_centres[i], 
                                           cluster_centres[j],
                                           nfeatures);
            if (tmp_dist < this_cluster_min_dist)
                this_cluster_min_dist = tmp_dist;
            /* find the min distance between i and centers > i */
        }
        min_dist += this_cluster_min_dist;
    }
    min_dist = SQR(min_dist/(float)(nclusters-1));
    ret = (float)(valid_sum/(npoints*min_dist) + penalty);

#else    /* modified by Wei-keng Liao */
    min_dist = FLT_MAX;
    for (i=0; i<nclusters; i++) {
        for (j=0; j<nclusters; j++) {
            if (i != j) {
                float dist;
                dist = euclid_dist_2(cluster_centres[i], 
                                     cluster_centres[j],
                                     nfeatures);
                if (dist < min_dist)
                    min_dist = dist;
            }
        }
    }
    ret = valid_sum/(npoints*min_dist) + penalty;
#endif

    return(ret);
}
